'use-strict';

const {convertStringObject} = require('../src/lib.js');

const	{value, toObject} = require('@clerksystems/concise-schema');

module.exports = (server) => {
	let api = null;
	function generateApi() {
		api = {};
		for (const endpointName in server.pluginComponents.endpoints) {
			api[endpointName] = {};
			const endpoints = server.pluginComponents.endpoints[endpointName].handlers;
			for (const methodName in endpoints) {
				const requestFormat = endpoints[methodName].requestFormat;
				if(typeof requestFormat === 'function' || requestFormat === undefined) {
					api[endpointName][methodName] = toObject(value.any);
				} else {
					api[endpointName][methodName] = toObject(convertStringObject(requestFormat));
				}
			}
		}
		return api;
	}
	//this should be called when something makes the requestFormat of a plugin change
	const apiNotifyRegister = {};
	server.requestFormatChanged = function () {
		generateApi();
		apiNotifyRegisterIds = Object.keys(apiNotifyRegister);
		if(apiNotifyRegisterIds.length === 0) {
			return;
		}
		for (const id of apiNotifyRegisterIds) {
			apiNotifyRegister[id](api);
		}
	}
	//plugin that exposes the api of all plugins
	return {
		name: 'api',
		endpoints: {
			api: {
				subscribe: {
					handler: function (connectionId, notify) {
						apiNotifyRegister[connectionId] = notify;
					}
				},
				unsubscribe: {
					handler: function (connectionId) {
						delete apiNotifyRegister[connectionId];
					}
				},
				read: {
					handler: function() {
						if(!api) {
							generateApi();
						}
						// TODO pick only endpoints that are available for the current connection.
						return api;
					}
				}
			}
		}
	};
}
