'use strict'

const   path                = require( 'path' )
const   uuid                = require( 'uuid' )
const	openSslParse        = require( 'opensslparser' )
const   _                   = require( 'lodash' )

const   lib                 = require( path.join( __dirname, 'lib.js' ) )
const	util				= require('util')

const	{validate, types}	= require('@clerksystems/concise-schema')

const   ConnectionWrapper   = function({
								connection,
								endpoints,
								staticEndpointArguments,
								pluginSettings,
								handlers,
								listeners,
								log,
								onClose,
								onError,
								pingTime = 10 * 1000,
								pingFailAmount = 6
							})
                            {
								// connection.psend = util.promisify(connection.send).catch( onError )
								// connection._X_send = connection.send
								// connection.send = function() {
								// 	connection.psend(...arguments).catch(onError)
								// }

								// connection.pping = util.promisify(connection.ping)
								// connection._X_ping = connection.ping
								// connection.ping = function() {
								// 	connection.pping(...arguments).catch(onError)
								// }

                                const   wrapper         = this
                                const   api             = {}
                                const   id              = uuid() // .substr( 0, 4 )

                                const   openedAt        = Date.now();

                                let     sessionData     = {}

                                const   remoteAddress   = connection.remoteAddress

                                try
                                {
                                    sessionData = openSslParse( connection )
                                }
                                catch( err ) {}

                                let     messageLog      = []

                                const   pluginHandles   = { setters: {}, getters: {} }      // moet private worden, clonen in de getter

                                const   store           = buildConnectionStore( pluginSettings, listeners, id )

                                const   storeHandlers   = {}

                                const   addToMessageLog = function( entry )
                                                        {
                                                            messageLog = [{ logTime: new Date().getTime(), ... entry }].concat( _.take( messageLog, 25 ) )
                                                        }

								const	sendMessage		= function( data ) {
															try
															{
																connection.send( JSON.stringify( data ))
															}
															catch ( err )
															{
																if( err.message.includes('readyState') ) {
																	log.debug( 'tried to send on closed connection' );
																} else {
																	console.error( 'Could not send message.', err.message );
																}
															}
														}

                                const   notify          = function( messageData )
                                                        {
                                                            const notification = makeNotifcationMessage( messageData )
                                                            addToMessageLog( notification )
                                                            try
                                                            {
                                                                connection.send( JSON.stringify( notification ))
                                                            }
                                                            catch ( err )
                                                            {
																if( err.message.includes('readyState') ) {
																	log.debug( 'tried tosend on closed connection' );
																} else {
																	console.error( 'Could not send notification message.', err.message );
																}
                                                            }
                                                        }

                                const   getter          = function( pluginName, key )
                                                        {
                                                            const val = _.get( store[ pluginName ], [ key, 'value' ], undefined )

                                                            if ( val === undefined )
                                                            {
                                                                return undefined
                                                            }
                                                            else if ( typeof val === 'object' )
                                                            {
                                                                //FIXME don't know about performance issues of this deep clone. but we need some form that preserves buffers
                                                                return _.cloneDeep(val)
                                                            }
                                                            return val
                                                        }

                                pluginHandles
                                    .getter = Object.freeze({ get: getter })

                                for ( const pluginName of Object.keys( store ) )
                                {
                                    const   setter          = function( key, val )
                                                            {
                                                                if ( _.get( store, [ pluginName, key ], undefined ) === undefined )
                                                                {
                                                                    const err = `Unknown setting referenced: ${pluginName}.${key}`
                                                                    log.warning( `Unknown setting referenced: ${pluginName}.${key}` )
                                                                    throw new Error( `Unknown setting referenced: ${pluginName}.${key}` )
                                                                }
                                                                store[ pluginName ][ key ].set( val )
                                                                return true
                                                            }

                                    pluginHandles
                                        .setters[ pluginName ] = Object.freeze({ set: setter })

                                }


                                // injectStoreSetter( store[ 'yoctopus' ] )





                                try
                                {
                                    log.debug( 'Calling onConnectionOpened plugin handlers' )

                                    const   handlerResult       = lib.callHandlers({
                                                                    handlers            : handlers.onConnectionOpened
                                                                ,   data                : { sessionData, remoteAddress }
                                                                ,   connectionId        : id
                                                                ,   pluginHandles
                                                                ,   notify
                                                                ,   staticEndpointArguments
                                                                })

                                    if ( handlerResult.errors )
                                    {
                                        for ( const error of handlerResult.errors )
                                        {
                                            log.error( `onConnectionOpened handler error: ${error}` )
                                        }

                                        log.info( 'Closing connection due to onConnectionOpened errors' )
                                        connection.close()
                                    }

                                    log.debug( 'Called onConnectionOpened plugin handlers' )
                                }
                                catch( err )
                                {
                                    log.error( `Could not complete calling connection.onOpen handlers on connection ${id}` )
                                    throw new Error( err )
                                }





                                connection.on( 'message', async function( msgStr )
                                {

                                    // prepare a new request log entry

                                    const requestMsgLogEntry = { msgStr }

                                    addToMessageLog( requestMsgLogEntry )


                                    // parse incoming JSON

                                    let msgObj = null

                                    try
                                    {
                                        msgObj = parseMessage( msgStr )
                                    }
                                    catch( err )
                                    {
                                        sendMessage( makeErrorMessage( _.get( msgObj, 'meta', {} ), `Request message parse error: ${err.message}` ) )
                                        return
                                    }


                                    // start processing request message by calling the first set of handlers

                                    log.debug( 'Calling onRequestReceived plugin handlers' )

                                    const   orrHandlerResult    = await lib.callHandlers({
                                                                    handlers            : handlers.onRequestReceived
                                                                ,   data                : msgObj
                                                                ,   connectionId        : id
                                                                ,   pluginHandles
                                                                ,   staticEndpointArguments
                                                                ,   notify
                                                                })

                                    requestMsgLogEntry
                                        .orrHandlerResult       = orrHandlerResult

                                    log.debug( 'Called onRequestReceived plugin handlers' )

                                    if ( orrHandlerResult.errors.length > 0 )
                                    {
                                        return sendMessage( makeErrorMessage( msgObj.meta, orrHandlerResult.errors ))
                                    }


                                    // find the requested endpoint and executes its handler

                                    log.debug( 'Calling endpoint handler' )

                                    const   response            = await handleRequest(
                                                                    orrHandlerResult.data
                                                                ,   endpoints
                                                                ,   staticEndpointArguments
                                                                ,   pluginHandles
                                                                ,   store
                                                                ,   log
                                                                ,   notify
                                                                ,   id
                                                                )

                                    requestMsgLogEntry
                                        .response               = response

                                    log.debug( 'Called endpoint handler' )


                                    // call downstream handlers

                                    log.debug( 'Calling onResponsePrepared plugin handlers' )

                                    const   orpHandlerResult    = await lib.callHandlers({
                                                                    handlers            : handlers.onResponsePrepared
                                                                ,   data                : response
                                                                ,   connectionId        : id
                                                                ,   pluginHandles
                                                                ,   staticEndpointArguments
                                                                ,   notify
                                                                })

                                    requestMsgLogEntry
                                        .orpHandlerResult       = orpHandlerResult

                                    log.debug( 'Called onResponsePrepared plugin handlers' )


                                    // respond

                                    log.debug( 'Sending response' )

                                    sendMessage( orpHandlerResult.data )



                                    // call after care handlers

                                    log.debug( 'Calling onResponseSent plugin handlers' )

                                    const   orsHandlerResult    = await lib.callHandlers({
                                                                    handlers            : handlers.onResponseSent
                                                                ,   data                : {}
                                                                ,   connectionId        : id
                                                                ,   pluginHandles
                                                                ,   staticEndpointArguments
                                                                ,   notify
                                                                })

                                    requestMsgLogEntry
                                        .orsHandlerResult       = orsHandlerResult

                                    log.debug( 'Called onResponseSent plugin handlers' )
                                })

								const closeHandler = async function()
                                {
									if( connection ) {
										connection.removeAllListeners()
									}

                                    log.info( `Connection closed (IP: ${remoteAddress})` )

                                    clearTimeout( pingTimeout )

                                    const   occHandlerResult    = await lib.callHandlers({
                                                                    handlers            : handlers.onConnectionClosed
                                                                ,   data                : {}
                                                                ,   connectionId        : id
                                                                ,   onError             : 'continue'
                                                                ,   pluginHandles
                                                                ,   staticEndpointArguments
                                                                })

                                    if ( occHandlerResult.errors.length > 0 )
                                    {
                                        for ( const error of occHandlerResult.errors )
                                        {
                                            log.error( `onConnectionClosed handler error: ${error}` )
                                        }
                                    }

                                    // log.debug( `onConnectionClosed handlers result ${JSON.stringify( occHandlerResult.data )}` )

                                    log.debug( 'Called onConnectionClosed plugin handlers' )

                                    onClose( id )
                                }

                                connection.on( 'close', closeHandler )
                                connection.on( 'error', closeHandler )



                                // ping logic to detect when a connection has silently closed

                                let     pingTimeout         = null
                                let     lostPings           = 0

                                const   resetPingTimeout    = function()
                                                            {
                                                                clearTimeout( pingTimeout )
                                                                pingTimeout = setTimeout( ping, pingTime )
                                                                lostPings = 0
                                                            }

                                const   ping                = function()
                                                            {
                                                                if ( lostPings >= pingFailAmount )
                                                                {
                                                                    log.notice( `Terminating connection to ${remoteAddress} after ${lostPings * pingTime / 1000}s of silence` )

                                                                    try
                                                                    {
                                                                        connection.terminate()

                                                                    }
                                                                    catch ( err )
                                                                    {
                                                                        log.error( `Error while terminating connection: ${err.message}` )
                                                                    }

                                                                    connection = null

                                                                    return
                                                                }
                                                                else
                                                                {
                                                                    if ( lostPings > 0 )
                                                                    {
                                                                        log.info( `Lost ${lostPings} ping to ${remoteAddress}` )
                                                                    }

																	if( connection ) {
																		connection.ping()
																	}

                                                                    lostPings += 1

                                                                    pingTimeout = setTimeout( ping, pingTime )
                                                                }
                                                            }

                                connection.on( 'pong', resetPingTimeout )
                                connection.on( 'message', resetPingTimeout )
                                connection.on( 'ping', resetPingTimeout )

                                resetPingTimeout()



                                Object.defineProperties(
                                    api
                                ,   {
                                        id: {
                                            get : () => id
                                        }
                                    ,   remoteAddress: {
                                            get: () => remoteAddress
                                        }
                                    ,   messageLog: {
                                            get: () => JSON.parse( JSON.stringify( messageLog ))        // deep clone?
                                        }
                                    ,   notify: {
                                            get: () => notify
                                        }
                                    ,   store: {
                                            get: () => exportStore( store )
                                        }
                                    ,   openedAt: {
                                            get: () => openedAt
                                        }
									,	close: {
											get: () => connection.close
										}
									,	lostPings: {
											get: () => lostPings
										}
                                    }
                                )

                                Object.freeze( api )

                                return api
                            }

const   exportStore             = store =>
                                    Object.entries( store )
                                        .reduce(
                                            ( exp, [ pluginName, pluginStore ] ) =>
                                                Object.assign(
                                                    exp
                                                ,   {
                                                        [ pluginName ]  : Object.keys( pluginStore )
                                                                            .reduce(
                                                                                ( expPluginStore, pluginSetting ) =>
                                                                                    Object.assign(
                                                                                        expPluginStore
                                                                                    ,   { [ pluginSetting ] : pluginStore[ pluginSetting ].value }
                                                                                    )
                                                                            ,   {}
                                                                            )
                                                    }
                                                )
                                        ,   {}
                                        )

                    // ({ store }) // remove setters - export only values

const   buildConnectionStore    = ( settingsObject, listenersObject, id ) =>
                                {
                                    const store = {}
                                    for ( let [ pluginName, settings ] of Object.entries( settingsObject ))
                                    {
                                        store[ pluginName ] = {}

                                        for ( let [ setting, settingType ] of Object.entries( settings ) )
                                        {
                                            const   listeners   = _.get( listenersObject, [ pluginName, setting ], [] )
                                                                    .concat( _.get( listenersObject, [ pluginName, '*' ], [] ) )
                                                                    .concat( _.get( listenersObject, [ '*' ], [] ) )

                                            const   settingPath = `${pluginName}.${setting}`

                                            store
                                                [ pluginName ]
                                                [ setting ]     = {
                                                                    value   : null
                                                                ,   set     : function( val )
                                                                            {
                                                                                if ( checkType( settingType, val ) !== true )
                                                                                {
                                                                                    throw new Error( `Connection ${id} setting "${setting}" requires a ${settingType} - rejecting value "${val}"` )
                                                                                }

                                                                                store[ pluginName ][ setting ].value = val

                                                                                if ( listeners.length > 0 )
                                                                                {
                                                                                    for ( let listener of listeners )
                                                                                    {
                                                                                        listener({ value: _.cloneDeep( val ), setting: settingPath, connectionId: id })
                                                                                    }
                                                                                }
                                                                            }
                                                                }
                                        }
                                    }

                                    Object.seal( store )

                                    return store
                                }


const   makeNotifcationMessage  = ({ id, type, data, customFields }) =>
                                ({
                                   ... customFields
                                ,   meta: {
                                        id          : id || uuid()
                                    ,   method      : 'notify'
                                    ,   type
                                    }
                                ,   data
                                })


const   makeErrorMessage    = ( meta, detail ) =>
                                ({
                                    meta: {
                                        id          : meta.id || uuid()
                                    ,   endpoint    : meta.endpoint
                                    ,   method      : meta.method
                                    }
                                ,   errors: Array.isArray( detail ) ? detail.map( err => ({ detail: err })) : [{ detail }]
                                })

const   makeDataMessage    = ( meta, data = {} ) =>
                                ({
                                    meta: {
                                        id          : meta.id || uuid()
                                    ,   endpoint    : meta.endpoint
                                    ,   method      : meta.method
                                    }
                                ,   data
                                })

const   checkRequirements   = async ( pluginStore, requirements, log, reqStr ) =>
                            {
                                for ( const [ reqPropPath, requiredValue ] of Object.entries( requirements ) )
                                {
                                    const actualValue = _.get( pluginStore, `${reqPropPath}.value`, undefined )

                                    if ( typeof ( requiredValue ) === 'function' )
                                    {
                                        try
                                        {
                                            const   calcValue   = await requiredValue( actualValue )

                                            if ( calcValue !== true )
                                            {
                                                log.error( `Unmet requirement fn at ${reqPropPath}: ( ${actualValue} ) => ${calcValue} is not "true" (${reqStr})` )
                                                return false
                                            }
                                        }
                                        catch( err )
                                        {
                                            log.error( `Requirement fn at ${reqPropPath}, threw an error on value "${actualValue}" (${reqStr})` )
                                            return false
                                        }
                                    }
                                    else if ( actualValue !== requiredValue )
                                    {
                                        log.error( `Unmet requirement: ${reqPropPath} is "${actualValue}" not "${requiredValue}" (${reqStr})` )
                                        return false
                                    }
                                }
                                return true
                            }

const   unknownType         = { validator: () => false }

const   checkType           = ( checkWith, val ) =>
                                (
                                    typeof ( checkWith ) === 'function'
                                    ?   checkWith
                                    :   _.get( types, checkWith, unknownType ).validator
                                )( val )

const   checkRequestFormat  = ( requestFormat, requestData = {} ) =>
                            {
								if(typeof requestFormat === 'function') {
									const errors = [];
									if(!requestFormat( requestData )) {
										errors.push( `Custom validation function rejected value ${JSON.stringify(requestData)}.` );
									}
									return errors;
								}
								return validate(requestFormat, requestData, false);
                            }


const   handleRequest       = async function( msgObj, endpoints, staticEndpointArguments, pluginHandles, pluginStore, log, notify, connectionId )
                            {
                                let     endpointObj         = null

                                try
                                {
                                    endpointObj             = findEPMObj( endpoints, msgObj.meta )
                                }
                                catch( err )
                                {
                                    return makeErrorMessage( msgObj.meta, 'Unknown method/endpoint combination' )
                                }



                                if ( endpointObj.requestFormat )
                                {
                                    const   formatErrors    = checkRequestFormat( endpointObj.requestFormat, msgObj.data )

                                    if ( formatErrors.length > 0 )
                                    {
                                        return makeErrorMessage( msgObj.meta, formatErrors )
                                    }
                                }


                                const   requirementsMet     = await checkRequirements( pluginStore, endpointObj.requirements || {}, log, msgObj.meta.method + ':' + msgObj.meta.endpoint )

                                if ( requirementsMet !== true )
                                {
                                    return makeErrorMessage( msgObj.meta, 'Request declined due to unmet endpoint requirements.' )
                                }
                                try
                                {
                                    return makeDataMessage(
                                        msgObj.meta
                                    ,   (
                                            await endpointObj
                                                .handler(
                                                {
                                                    ...staticEndpointArguments
                                                ,   connectionId
                                                ,   request             : msgObj
                                                ,   connectionSettings  : endpointObj.handleScopeBuilder( pluginHandles )
                                                ,   log                 : endpointObj.logger
                                                ,   notify
                                                })
                                        )
                                    )
                                }
                                catch( err )
                                {
                                    console.log( err )
                                    return makeErrorMessage( msgObj.meta, `Internal error while processing request: ${err.message}` )
                                }
                            }


const   findEPMObj          = ( endpoints, requestMeta ) =>
                            {
                                const endpointObject = endpoints[ requestMeta.method + ':' + requestMeta.endpoint ]

                                if ( typeof ( endpointObject ) !== 'object' )
                                {
                                    throw new Error( 'Unkown method/endpoint combination: ' + requestMeta.endpoint )
                                }

                                return endpointObject
                            }


const   parseMessage        = msgStr =>
                            {
                                let msgObj = {}

                                try
                                {
                                    msgObj = JSON.parse( msgStr )
                                }
                                catch( err )
                                {
                                    throw new Error( 'This is not JSON: ' + msgStr )
                                }

                                if ( typeof msgObj.meta !== 'object' )
                                {
                                    throw new Error( 'Missing the required meta entry' )
                                }
                                if ( typeof msgObj.meta.id !== 'string' )
                                {
                                    throw new Error( 'Missing the required meta.id entry' )
                                }

                                if ( typeof msgObj.meta.endpoint !== 'string' )
                                {
                                    throw new Error( 'Missing the required meta.endpoint entry' )
                                }

                                if ( typeof msgObj.meta.method !== 'string' )
                                {
                                    throw new Error( 'Missing the required meta.method entry' )
                                }

                                return msgObj
                            }


module.exports              = ConnectionWrapper
