'use strict';

const {value, types} = require('@clerksystems/concise-schema');

const   exp         = {}

module.exports      = exp

exp.callHandlers    = async ({ handlers, data, staticEndpointArguments, notify, pluginHandles, socket, connectionId, onError = 'return'}) =>
                        {
                            const errors = []

                            for ( let handler of handlers )
                            {
                                try
                                {
                                    let newData   = await handler.fn({
                                                            connectionSettings  : handler.handleScopeBuilder( pluginHandles )
                                                        ,   log                 : handler.logger
                                                        ,   data
                                                        ,   ...staticEndpointArguments
                                                        ,   notify
                                                        ,   socket
                                                        ,   connectionId
                                                        })

                                    if ( newData )      // K: dit is niet per se slim
                                    {
                                        data = newData
                                    }
                                }
                                catch( err )
                                {
                                    console.log( err )

                                    errors.push( `Handler ${handler.name} error: ${err.message}` )

                                    switch ( onError )
                                    {
                                        case 'throw':
                                            throw new Error( err );

                                        case 'return':
                                            continue;

                                        case 'continue':
                                            // do nothing
                                            break;
                                    }
                                }
                            }

                            return { errors, data }
                        }



exp.enableDestroy       = function(server) {
                            var connections = {}

                            server.on('connection', function(conn) {
                                var key = conn.remoteAddress + ':' + conn.remotePort;
                                connections[key] = conn;
                                conn.on('close', function() {
                                	delete connections[key];
                                });
                            });

                            server.destroy = function(cb) {
                                server.close(cb);
                                for (var key in connections) {
                                	connections[key].destroy();
									delete connections[key];
								}
                            }
                        }

const	normalizeSchema	= schema =>
						{
							if ( schema !== undefined && ! ( schema.isVV === true ) )
							{
								if ( types.object.validator( schema ) === true )
								{
									schema = value.required.object.like( schema );
								}
								else if ( types.array.validator( schema ) === true )
								{
									schema =
										schema.length === 1
										?	value.required.array.like( schema[ 0 ] )
										:	value.required.array
										;
								}
								else
								{
									const e = new Error('Unknown expression: [' + schema + ']');
									console.log(e.stack);
									throw e;
								}
							}

							return schema
						}
exp.convertStringObject	= function(schema)
						{
							if( schema.isVV ) {
								return schema;
							}

							if(typeof schema === 'string')
							{
								schema = value[schema].required;
							}
							else
							{
								schema = normalizeSchema(schema);
								if(schema !== undefined && schema.subValue) {
									if( types.array.validator( schema ) === true )
									{
										schema.subValue = exp.convertStringObject(schema.subValue)
									}
									else
									{ // object
										for (const key in schema.subValue) {
											schema.subValue[key] = exp.convertStringObject(schema.subValue[key])
										}
									}
								}
							}

							return schema;
						}
